import functools as fc
import asyncio
import logging
import math
import codecs
import ujson as json
from networktools.time import timestamp
from networktools.colorprint import gprint, bprint, rprint
from data_amqp import AMQPData
from tasktools.taskloop import TaskLoop


"""
La implementación de una clase heredera

Signatures:
==========

- a :: array
- s :: string
- v :: variant
- a{sv} :: un diccionario

ref:: https://www.gkbrk.com/2018/02/simple-dbus-service-in-python/
"""


"""

DBusConnection
"""


class RMQEngine:
    def __init__(self, *args, **kwargs):
        self.__status = False
        self.amqp = None
        if kwargs.get('amqp', False):
            try:
                self.amqp_opt = kwargs['amqp']
                self.amqp = AMQPData(**kwargs)
                self.__status = True
            except Exception as e:
                self.__status = False
                raise e

    @property
    def status(self):
        return self.__status

    def enu_data(self, msg):
        try:
            self.amqp.manage_json_data(msg)
        except Exception as e:
            raise e

    def solo_data(self, msg):
        try:
            self.amqp.manage_json_data(msg.get('data', {}))
        except Exception as e:
            raise e


class DataManager:
    def __init__(self, queue, engine_opts, step=0.1, *args, **kwargs):
        self.step = step
        self.queue = queue
        self.engine_opts = engine_opts
        print("Channels", list(engine_opts.keys()))
        self.engine = False

    def create_rmq(self, channel, engine):
        try:
            value_dict = self.engine_opts.get(channel, {})
            print("Input to RMWEngine")
            rmq = RMQEngine(**value_dict)
            engine.update({channel: rmq})
            return rmq
        except Exception as e:
            print("Can't create RMQ connection %s" % e)
            raise e

    async def cycle(self, *args, **kwargs):
        engine = args[0]
        v = args[1]
        init_set = args[2]
        queue = self.queue
        await asyncio.sleep(1)
        if v == 0:
            engine = {}
            try:
                added_set = set()
                for channel in init_set:
                    print("Init conection to channels rmq", channel)
                    value_dict = self.engine_opts.get(channel)
                    rmq = RMQEngine(**value_dict)
                    engine.update({channel: rmq})
                    added_set.add(channel)
                for item in added_set:
                    init_set.remove(item)
                v += 1
            except Exception as e:
                print("Error al intentar conectar rabbitmq %s" % e)
                v = 0
            return [engine, v, init_set], kwargs
        else:
            if engine:
                if not queue.empty():
                    for i in range(queue.qsize()):
                        msg = queue.get()
                        if type(msg) == dict:
                            # get channel from msg, to direct the msg
                            channel = msg.get('channel', None)
                            if channel in engine:
                                try:
                                    if engine.get(channel).status:
                                        # select the dbus channel and send data
                                        if channel == 'earlybird':
                                            engine.get(channel).solo_data(msg)
                                        else:
                                            engine.get(channel).enu_data(msg)
                                except Exception as ex:
                                    print("Error en comunicar %s, channel %s" %
                                          (ex, channel))
                                    engine.get("amqp").connect()
                                    init_set.add(channel)
                                    v = 0
                        queue.task_done()
            await asyncio.sleep(self.step)
            return [engine, v, init_set], kwargs

    def task_cycle(self):
        loop = asyncio.get_event_loop()
        # engine = {channel:
        #          for channel, value_dict in self.engine_opts.items()}
        rprint("#"*30)
        bprint("Task Cycle Ok")
        rprint("#"*30)
        v = 0
        init_set = set(self.engine_opts.keys())
        task = TaskLoop(self.cycle, [{}, v, init_set],
                        {}, **{"name": "task_data_manager"})
        task.create()
        if not loop.is_running():
            loop.run_forever()



import socket
from datetime import datetime, date
from networktools.path import home_path
import asyncio
from asyncio import wait_for
from pathlib import Path
# NetworkTools methods and classes:
from data_rdb import Rethink_DBS
from data_geo import GeoJSONData
from data_amqp import AMQPData


from networktools.library import my_random_string, check_gsof
from networktools.colorprint import gprint, bprint, rprint
from networktools.geo import (radius, deg2rad,
                              ecef2llh, llh2ecef)
from networktools.time import get_datetime_di
from networktools.library import geojson2json

# log module
from basic_logtools.filelog import LogFile

# TaskTools for concurrency loop
from tasktools.taskloop import TaskLoop
from tasktools.scheduler import TaskScheduler

# Async Socket
# from gnsocket.gn_socket import GNCSocket


# Maths
import time
from multiprocessing import Lock
import json

# RethinkDB module
from rethinkdb import RethinkDB
rdb = RethinkDB()
rdb.set_loop_type('asyncio')


def default(o):
    if isinstance(o, (date, datetime)):
        return o.isoformat()


def qjson(x):
    return json.loads(
        json.dumps(x,
                   sort_keys=True,
                   indent=1,
                   default=default)
    )


class DragonGather(TaskScheduler):
    log_manager = {}

    def __init__(self, queue_list, *args, **kwargs):
        # super().__init__()

        self.key = 'DT_GEN'
        self.ipt = kwargs.get('ipt')
        self.ico = kwargs.get('ico')
        self.process_queue = kwargs.get('process_queue')
        self.gui_queue = kwargs.get('gui_queue')
        self.proc_tasks = kwargs.get("proc_tasks")
        self.queue_client_n2t = kwargs.get("queue_client_n2t")
        self.queue_client_t2n = kwargs.get("queue_client_t2n")
        self.queue2client = kwargs.get("queue2client")
        self.stations = kwargs.get("stations")
        self.position = kwargs.get("position")
        self.db_data = kwargs.get("db_data")
        self.lnproc = kwargs.get("nproc")
        self.sta_init = kwargs.get("sta_init")
        self.bridge = kwargs.get("bridge")
        self.common = kwargs.get("common")
        self.rethinkdb_address = kwargs.get('rdb_address')
        self.rethinkdb_enu_address = kwargs.get('rdb_enu_address')
        self.rethinkdb_dbname = kwargs.get('rdb_dbname', 'test')
        self.sta_init = kwargs.get('status_tasks')
        self.status_tasks = kwargs.get('status_tasks')
        self.assigned_tasks = kwargs.get('assigned_tasks')
        self.group = kwargs.get('group')
        self.gui_group = kwargs.get('gui_group')
        self.gui_set = kwargs.get('gui_set')
        self.signals = kwargs.get('signals')
        self.isg = kwargs.get('isg')
        self.sigid = kwargs.get('sigid')
        self.sc = kwargs.get('send_control')
        self.dbus_queue = kwargs.get('dbus_queue')
        self.dbus_geojson_queue = kwargs.get('dbus_geojson_queue')
        self.status_keys = kwargs.get('status_keys', {'BATT_MEM', 'DOP'})
        args = []
        # call superior class
        super().__init__(*args, **kwargs)
        #
        self.enqueued = kwargs.get('enqueued')
        self.rq = queue_list[0]
        self.wq = queue_list[1]
        self.queue_list = queue_list
        self.start = 0
        self.signal = None
        self.origin_objects = {"RethinkDB": Rethink_DBS}
        self.rethinkdb = None
        self.rethinkdb_origin = {}
        self.rethinkdb_enu_origin = {}
        self.lock = None
        self.msg_process = dict(
            GET_LST={'STATION': self.load_stations,
                     'DBDATA': self.load_dbdata}
        )
        self.msg_server_process = dict(
            GET_STA=self.get_sta
        )

        coros_callback_dict = {
            'run_task': self.gather_data,
        }
        self.set_new_run_task(**coros_callback_dict)
        self.log_path = Path(home_path(kwargs.get('log_path', '~/log')))

    @property
    def class_name(self):
        return self.__class__.__name__

    def set_isg(self, uin=4):
        """
        Defines a new id for task related to collect data insice a worker,
        check if exists
        """
        isg = my_random_string(uin)
        while True:
            if isg not in self.isg:
                self.isg.append(isg)
                break
            else:
                isg = my_random_string(uin)
        return isg

    def msg_network_task(self):
        # get from queue status from SOCKET
        # send result
        # read_queue -> rq
        # process msg -> f(msg)
        queue_list = [self.rq, self.wq]
        time.sleep(3)
        loop = asyncio.get_event_loop()
        self.load_data_task(loop)
        # gprint("XDX Gestionando mensajes en engine", flush=True)
        try:
            args = []
            # Create instances

            task_client = TaskLoop(self.check_client_status, args, {},
                                   **{"name": "check_client_status"})

            task = TaskLoop(self.check_status, args, {},
                            **{"name": "check_status"})

            task_client.create()
            task.create()

            if not loop.is_running():
                loop.run_forever()
        except Exception as ex:
            print("Error o exception que se levanta con %s" %
                  format(queue_list))
            print(ex)
            raise ex

    async def check_client_status(self, *args, **kwargs):
        q2client = self.queue2client
        wq = self.queue_client_n2t
        rq = self.queue_client_t2n
        await asyncio.sleep(.5)
        try:
            # first read the msgs from system and send to client
            # check if is dict
            # check if has the main keys....(?)
            if not q2client.empty():
                for i in range(q2client.qsize()):
                    msg = q2client.get()
                    if isinstance(msg, dict):
                        wq.put(msg)
            # read the queue that receive the data on client
            # process the msg
            # do something...
            if not rq.empty():
                for i in range(rq.qsize()):
                    msg = rq.get()
                    result = await self.client_interpreter(msg)
                    rprint("Resultado recibido de client:%s" % result)
                    """
                    if not None:
                        wq.put({'msg': result,
                                'idc': idc})
                    """
            # bprint(self.instances.keys())
        except Exception as ex:
            print(ex)
            raise ex
        return args, kwargs

    async def client_interpreter(self, msg_in):
        # msg is a string JSON
        self.lock = Lock()
        result = []
        msg = msg_in.get('dt').get('msg')
        print("Msg recv")
        if isinstance(msg, dict):
            command = msg.get('command').get('action')
            answer = msg.get('command').get('answer')
            result = {}
            gprint("*"*20)
            bprint(answer.keys())
            gprint("*"*20)
            if command in self.msg_process and answer:
                rprint("Creating...")
                varname = msg.get('command').get('varname')
                def print_dict(x): return [print(
                    "%s->%s" % (k, v) for k, v in x.items())]
                msg_process = self.msg_process.get(
                    command, {}).get(varname, print_dict)
                bprint("msg process _> %s" % msg_process)
                result = msg_process(answer)
                rprint(f"Result from process list {result}")
        return result

    async def check_status(self, *args, **kwargs):
        wq = self.queue_list[0]
        rq = self.queue_list[1]
        await asyncio.sleep(.5)
        try:
            if not rq.empty():
                for i in range(rq.qsize()):
                    msg = rq.get()
                    try:
                        m = msg.get('dt')
                        idc = msg.get('idc')
                        result = await self.interpreter(m)
                        if result:
                            wq.put({'msg': result,
                                    'idc': idc})
                    except Exception as ex:
                        print("Error al transformar")
                        print(ex)
                        print(msg)
                        print(type(msg))
                        raise ex
            # bprint(self.instances.keys())
        except Exception as ex:
            print(ex)
            raise ex
        return args, kwargs

    async def interpreter(self, msg):
        # msg is a string JSON
        self.lock = Lock()
        command = msg.get('command', 'GET_STA')
        args = msg.get('args', [])
        result = None
        if command == 'init_gui':
            # if self.gui_group empty:
            if not self.gui_group:
                with self.lock:
                    self.sc.value = not self.sc.value
            # add group and id
            idg = args[0]
            group = args[1]
            # relationship between a gui and a group of stations
            self.gui_group.update({idg: group})

        if command in self.msg_server_process:
            result = self.msg_server_process.get(command)(*args)
        return result

    def get_sta(self, *answer):
        stations = self.stations
        # gprint("XDX stations keys : %s" %stations.keys())
        for ids in stations.keys():
            # bprint("XDX Cargando data a process")
            dataset = stations[ids]
            try:
                if dataset.get('code') in self.group or 'ALL' in self.group:
                    # if self.sc.value:
                    self.dbus_queue.put({
                        'command': 'station',
                        'data': {ids: self.stations[ids]}})
                    self.dbus_queue.put({
                        'command': 'position',
                        'data': {ids: self.position[ids]}})
                    self.gui_set.append(dataset['code'])
                    # print("XDX .sds", flush=True)

                # gprint(self.queue_process.qsize())
            except Exception as exc:
                print("Error al cargar ids a queue")
                raise exc
        self.dbus_queue.put({
            'channel': 'data',
            'command': 'load_chart',
            'data': []})
        return json.dumps({
            'channel': 'data',
            'command': 'GET_STA',
            'args': ['dbus']
        })

    def add_process_queue(self, process_queue):
        self.process_queue = process_queue

    def load_stations(self, msg):
        qs = msg
        POSITION = dict()
        for ids in qs:
            qs[ids]['STATUS'] = 'OFF'
            POSITION[ids] = dict()
            POSITION[ids]['ECEF'] = dict()
        for ids in qs:
            self.stations[ids] = qs.get(ids)
            print("Station", qs.get(ids))
            if 'ECEF_Z' in qs[ids]:
                Z = qs[ids]['ECEF_Z']
                POSITION[ids]['ECEF'].update({'Z': Z})
            if 'ECEF_X' in qs[ids]:
                X = qs[ids]['ECEF_X']
                POSITION[ids]['ECEF'].update({'X': X})
            if 'ECEF_Y' in qs[ids]:
                Y = qs[ids]['ECEF_Y']
                POSITION[ids]['ECEF'].update({'Y': Y})
            if 'position' in qs[ids]:
                pst = json.loads(qs[ids]['position'])
                coords = pst['coordinates']
                [lat, lon] = deg2rad(*coords)
                POSITION[ids].update({'lat': lat})
                POSITION[ids].update({'lon': lon})
                POSITION[ids].update({'radius': radius(lat)[0]})
                XYZ = llh2ecef(lat, lon, Z)
                # bprint(XYZ) ok, correct
                POSITION[ids].update({'ECEF': dict(zip(['X', 'Y', 'Z'], XYZ))})
            x = POSITION[ids]['ECEF']['X']
            y = POSITION[ids]['ECEF']['Y']
            z = POSITION[ids]['ECEF']['Z']
            (lat, lon, h) = ecef2llh(x, y, z)
            POSITION[ids].update({'llh': {'lat': lat, 'lon': lon, 'z': h}})
            self.position[ids] = POSITION[ids]
            if 'code' in qs:
                self.common[qs][ids] = dict()
            dataset = self.stations.get(ids)
            # code = dataset.get('code')
            self.process_queue.put(ids)
            self.enqueued.append(ids)
            self.gui_set.append(dataset['code'])
            for k, s in self.stations.items():
                bprint(f"{k} : {s}")

    def load_dbdata(self, msg):
        for k, v in msg.items():
            self.db_data[k] = v

    def msg_load_stations(self):
        get_lst = {'user': 'admin',
                   'group': 'admin',
                   'command': {'action': 'GET_LST',
                               'varname': 'STATION'}}
        return get_lst

    def msg_load_dbdata(self):
        get_lst = {'user': 'admin',
                   'group': 'admin',
                   'command': {'action': 'GET_LST',
                               'varname': 'DBDATA'}}
        return get_lst

    def init_datawork_data(self):
        rprint("Inicializando datawork data")
        msg_list = [self.msg_load_stations(), self.msg_load_dbdata()]
        for msg in msg_list:
            print("Mensaje ->", msg)
            self.queue2client.put(msg)

    def activate_stations(self, stations):
        for ids in stations.keys():
            dataset = stations.get(ids)
            self.activate_station(dataset, ids)

    def activate_station(self, dataset, ids):
        try:
            if dataset['code'] in self.group:
                self.process_queue.put(ids)
                self.gui_set.append(dataset['code'])
        except Exception as exc:
            print("Error al cargar ids a queue")
            raise exc

    async def load_data(self, *args, **kwargs):
        """
        Load main data at the beggining
        In the future, must handle messages betwen
        DragonDataWork and Collector

        """
        if self.start == 0:
            print("Enviando msg inicial")
            self.init_datawork_data()
            self.start = 1
            ###
        else:
            await asyncio.sleep(25)
        return args, kwargs

    def load_data_task(self, loop):
        # bprint("Load data task")
        args = []
        task = TaskLoop(self.load_data, args, {}, **{"name": "load_data_task"})
        task.create()

    def add_sta_instance_origin(self, ids, loop):
        # create bridge instance
        # bprint("Station %s and port %s" % (self.stations[ids]['code'], self.bridge))
        # gprint("Bridge: %s" % format(self.bridge))
        # local_port = self.create_bridge(ids)
        # rprint("Local port %s" %local_port)
        code = self.stations[ids]['code']
        code_db = self.stations[ids]['db']
        # idd = self.get_id_by_code('DBDATA', code_db)
        # bprint("The db_data's id: %s" % idd)
        print("Conexión source", self.rethinkdb_address,
              f"DBNAME -> {self.rethinkdb_dbname}")
        db_datos = dict(host='localhost',
                        name="Source_%s" % code_db,
                        code=code_db,
                        port=self.rethinkdb_address[1],
                        address=self.rethinkdb_address,
                        dbname=self.rethinkdb_dbname,
                        io_loop=loop,
                        log_path=str(self.log_path/"source_rdb"),
                        env='gather_%s' % code)
        name = 'RethinkDB'
        self.sta_init[ids] = True
        if name == 'RethinkDB':
            self.rethinkdb_origin[ids] = True
        try:
            RDB_C = self.origin_objects[name]
            rethinkdb_c = RDB_C(**db_datos)
            time.sleep(.5)
            return rethinkdb_c
        except Exception as ex:
            print("Error al inicializar conexión %s" % ex)
            raise ex

    def add_sta_instance_destiny(self, ids, loop):
        # create bridge instance
        # bprint("Station %s and port %s" % (self.stations[ids]['code'], self.bridge))
        # gprint("Bridge: %s" % format(self.bridge))
        # local_port = self.create_bridge(ids)
        # rprint("Local port %s" %local_port)
        code = self.stations[ids]['code']
        code_db = self.stations[ids]['db']
        # idd = self.get_id_by_code('DBDATA', code_db)
        # bprint("The db_data's id: %s" % idd)
        bprint(f"Iniciando RDB para {code} en {code_db}")
        print("Conexión enu", self.rethinkdb_enu_address)
        db_datos_enu = dict(host='localhost',
                            name="Destiny_enu",
                            code=code_db,
                            port=self.rethinkdb_enu_address[1],
                            address=self.rethinkdb_enu_address,
                            dbname='enu_data',
                            io_loop=loop,
                            log_path=str(self.log_path/"enu_rdb"),
                            env='enu_%s' % code)

        gprint("Data----->")
        rprint(db_datos_enu)
        gprint("<-----Data")
        name = 'RethinkDB'
        self.sta_init[ids] = True
        if name == 'RethinkDB':
            self.rethinkdb_enu_origin[ids] = True
        try:
            RDB_C = self.origin_objects[name]
            rethinkdb_c_enu = RDB_C(**db_datos_enu)
            bprint(f"Objeto->{rethinkdb_c_enu}")
            time.sleep(.5)
            return rethinkdb_c_enu
        except Exception as ex:
            print("Error al inicializar conexión %s" % ex)
            raise ex

        # self.common[code] = dict()
    # CREATE QUEUE INSTANCES

    # GET DATA AND SEND TO PLOT
    #
    #

    def add_process_instance(self, ids):
        CODE = self.stations[ids]['protocol'].upper()
        station = self.stations[ids]['code']
        kwargs = dict()
        kwargs['code'] = CODE
        kwargs['station'] = self.stations[ids]
        kwargs['position'] = self.position[ids]
        kwargs['log_path'] = str(self.log_path/'geo_json_data')
        process_instance = GeoJSONData(**kwargs)
        return process_instance

    def add_ew_instance(self, ids):
        ew_host = '10.54.218.81'
        creds = ('seismic', 'secret')
        CODE = self.stations[ids]['protocol'].upper()
        station = self.stations[ids]['code']
        kwargs = dict()
        kwargs['code'] = CODE
        kwargs['station'] = self.stations[ids]
        kwargs['position'] = self.position[ids]
        kwargs['host'] = ew_host
        kwargs['credentials'] = creds
        ew_instance = AMQPData(**kwargs)
        return ew_instance

    async def gather_data(self, ipt, ids, *args, **kwargs):
        """
        This method it's maybe the most important because generate
        the instances and gather the data in a general loop

        The logging system is a task by process
        """
        # input : ids, loop, sta
        uin = 3
        idi = my_random_string(uin)
        loop = asyncio.get_event_loop()
        v = int(args[0])
        sta = args[1]
        di, control, (rc, rc_enu), process_data = sta
        sta_init_flag = self.sta_init.get(ids)
        code = self.stations[ids]['code']
        code_db = self.stations[ids]['db']
        log = kwargs.get('log')
        # self.rethinkdb_origin[ids] = True
        # self.rethinkdb_enu_origin[ids] = True
        """
        Objetos centrales, de operacion
        """
        body, origin, destiny = False, False, False
        await asyncio.sleep(.5)

        if not sta_init_flag:
            # para controlar que?
            control = None
            # punto de inicio a consultar
            di = rdb.iso8601(get_datetime_di(delta=30))
            # el operador o procesador de la info
            process_data = self.add_process_instance(ids)
            self.sta_init[ids] = True
            body = True

        """
        Iniciar objeto database fuente
        """
        if not self.rethinkdb_origin.get(ids):
            # las dos instancias a dbs rethinkdb
            # rc: origen
            # rec_enu: destino
            rc = self.add_sta_instance_origin(ids, loop)
            try:
                await wait_for(rc.async_connect(), timeout=120)
                rc.set_defaultdb(self.rethinkdb_dbname)
                await rc.list_dbs()
                await rc.select_db(self.rethinkdb_dbname)
                table_name = self.stations[ids]['db']  # created on datawork
                indexes = await rc.get_indexes(table_name)
                await rc.list_tables(rc.default_db)
                origin = True
            except asyncio.CancelledError as ce:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TOA_-2 + {idex} + {code} +
                {ce}, ConnectionClosed rc_enu connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except (socket.error,
                    ConnectionResetError,
                    ConnectionAbortedError) as conn_error:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TOA_-1 + {idex} + {code} +
                {conn_error}, ConnectionClosed rc_enu connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except asyncio.TimeoutError as e:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f"PD_TO1_00 + {idex} + {code} + {e}"
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                rc.close()
                del rc
                rc = None
                return [ipt, ids, v,
                        [di, control, (rc, rc_enu),
                         process_data]], kwargs
        """
        Iniciar objeto database destino
        """
        if not self.rethinkdb_enu_origin.get(ids):
            # las dos instancias a dbs rethinkdb
            # rc: origen
            # rec_enu: destino
            rc_enu = self.add_sta_instance_destiny(ids, loop)
            try:
                await wait_for(rc_enu.async_connect(), timeout=120)
                await rc_enu.list_dbs()
                await rc_enu.select_db('enu_data')
                table_name = self.stations[ids]['db']  # created on datawork
                await rc_enu.create_table(table_name, rc_enu.default_db)
                if 'DT_GEN' not in indexes:
                    await rc_enu.create_index(table_name, index='DT_GEN')
                indexes = await rc_enu.get_indexes(table_name)
                await rc_enu.list_tables(rc_enu.default_db)
                destiny = True
            except asyncio.CancelledError as ce:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TOB_-2 + {idex} + {code} +
                {ce}, ConnectionClosed rc_enu connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except (socket.error,
                    ConnectionResetError,
                    ConnectionAbortedError) as conn_error:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TOB_-1 + {idex} + {code} +
                {conn_error}, ConnectionClosed rc_enu connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except asyncio.TimeoutError as e:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f"PD_TO1_00 + {code} + {e}"
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin_enu[ids] = False
                bprint(kwargs)
                rc_enu.close()
                del rc_enu
                rc_enu = None
                return [ipt, ids, v,
                        [di, control, (rc, rc_enu),
                         process_data]], kwargs
        if body or origin or destiny:
            v = 2
            wargs = [ipt, ids, v, [di, control, (rc, rc_enu), process_data]]
            return wargs, kwargs
        elif sta_init_flag:
            code = self.stations[ids]['code']
            table_name = self.stations[ids]['db']
            key = self.key
            try:
                df = rdb.iso8601(get_datetime_di(delta=0))  # now
                filter_opt = {'left_bound': 'open', 'index': key}
                future = rc.get_data_filter(table_name,
                                            [di, df],
                                            filter_opt,
                                            key)
                cursor = await wait_for(future, timeout=240)
            except asyncio.CancelledError as ce:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TOA_-2 + {idex} + {code} +
                {ce}, ConnectionClosed rc_enu connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except (socket.error,
                    ConnectionResetError,
                    ConnectionAbortedError) as conn_error:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f'''PD_TO3_-1 + {idex} + {code} +
                {conn_error}, ConnectionClosed rc connection....'''
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except asyncio.TimeoutError as e:
                idex = my_random_string(uin)
                kwargs["origin_exception"] = f"PD_TO3_00 + {idex} + {code} + {e}"
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                bprint(kwargs)
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs
            except Exception as ex:
                idex = my_random_string(uin)
                print("IDEX %s Exception" % idex, ex)
                print("Error en obtención de data desde rethinkdb")
                rc.logerror(
                    f'''IDEX = {idex}, Error en la obtención de datos para
                    estación {code} en {ex}''')
                kwargs["origin_exception"] = f"PD_TO4_00 + {code} + {idex} + {ex}"
                log.exception(kwargs["origin_exception"])
                self.rethinkdb_origin[ids] = False
                rc.close()
                del rc
                rc = None
                wargs = [ipt, ids, v, [
                    di, control, (rc, rc_enu), process_data]]
                return wargs, kwargs

            # signal.message({"msg":"cursor len %s" %len(cursor)})
            list_ok = []
            for c in cursor:
                if check_gsof(c):
                    try:
                        result = process_data.manage_data(c)
                        # to_plot = geojson2json(result)
                        to_db = geojson2json(result, destiny='db')
                        # result_enu =
                        try:
                            await wait_for(
                                rc_enu.save_data(table_name, to_db),
                                timeout=60)
                        except asyncio.CancelledError as ce:
                            idex = my_random_string(uin)
                            kwargs["origin_exception"] = f'''PD_TO3_-2 + {idex} + {code} +
                            {ce}, ConnectionClosed rc_enu connection....'''
                            log.exception(kwargs["origin_exception"])
                            self.rethinkdb_origin[ids] = False
                            bprint(kwargs)
                            rc_enu.close()
                            del rc_enu
                            rc_enu = None
                            wargs = [ipt, ids, v, [
                                di, control, (rc, rc_enu), process_data]]
                            return wargs, kwargs
                        except (socket.error,
                                ConnectionResetError,
                                ConnectionAbortedError) as conn_error:
                            idex = my_random_string(uin)
                            kwargs["origin_exception"] = f'''PD_TO3_-1 + {idex} + {code} + {conn_error}, ConnectionClosed rc_enu connection....'''
                            log.exception(kwargs["origin_exception"])
                            self.rethinkdb_origin[ids] = False
                            bprint(kwargs)
                            rc.close()
                            del rc
                            rc = None
                            wargs = [ipt, ids, v, [
                                di, control, (rc, rc_enu), process_data]]
                            return wargs, kwargs
                        except Exception as e:
                            bprint(f"Algo pasa con rc_enu....{e}")
                        # send normal json
                        msg_result = qjson(result)
                        send = {
                            'channel': 'earlybird',
                            'command': 'add_data',
                            'data': msg_result}
                        self.dbus_queue.put(send)
                        # send to earlybird
                        await asyncio.sleep(.1)
                        send_eb = {
                            'channel': 'earlybird',
                            'command': 'add_data',
                            'data': msg_result}
                        try:
                            self.dbus_queue.put(send_eb)
                        except Exception as e:
                            rprint("Error al enviar send_ev")
                            bprint(e)
                        intersection = self.status_keys.intersection(c.keys())
                        if intersection:
                            data_msg = {key: qjson(c.get(key))
                                        for key in intersection}
                            send = {
                                'channel': 'status',
                                'command': 'add_status',
                                'station': code,
                                'data': data_msg}
                            self.dbus_queue.put(send)
                        if self.dbus_geojson_queue:
                            self.dbus_geojson_queue.put(send)
                    except asyncio.TimeoutError as e:
                        kwargs["origin_exception"] = f"PD_TO5_01 + {code} + {e}"
                        log.exception(kwargs["origin_exception"])
                        self.rethinkdb_origin_enu[ids] = False
                        rc_enu.close()
                        del rc_enu
                        rc_enu = None
                        wargs = [ipt, ids, v, [
                            di, control, (rc, rc_enu), process_data]]
                        return wargs, kwargs
                    except Exception as ex:
                        print("WS Error al enviar %s a cola %s" %
                              (code, ex), flush=True)
                        kwargs["origin_exception"] = "PD_TO5_02 + %s" % code
                        log.exception(kwargs["origin_exception"])
                        bprint(kwargs)
                        rc_enu.close()
                        del rc_enu
                        rc_enu = None
                        return [ipt, ids, v,
                                [di, control,
                                 (rc, rc_enu),
                                 process_data]], kwargs

                    list_ok.append(c)
                else:
                    msg = "Error en estacixon %s  ----> %s" % (code, c)
                    await rc.msg_log(msg, "DEBUG")
            if cursor:
                dt_recv = cursor[-1].get('DT_RECV')
                di = cursor[-1].get('DT_GEN', dt_recv)
                control = True
            return [
                ipt, ids, v,
                [di, control, (rc, rc_enu), process_data]
            ], kwargs
        return [ipt, ids, v, sta], kwargs

    def set_pst(self, ids, args, kwargs):
        return [args[0], ids, *args[1:]], kwargs

    def set_init_args_kwargs(self, ipt):
        """
        This definition is for collector instance
        """
        rprint(f"Init log..... for ipt {ipt}")
        log = print
        log = LogFile("Engine@Datawork", "CORE_%s" %
                      ipt, "localhost@pineiden",
                      path=str(self.log_path/"engine"))
        self.log_manager[ipt] = log
        return [ipt, 1, (None, None, (None, None), None)], {"log": log}



# standar library
import os
import sys
import subprocess
import socket
import shlex
import asyncio
import concurrent.futures
import time
import functools
import math
import argparse
from ctypes import c_bool
from multiprocessing import Manager, Queue, Lock
from multiprocessing.managers import BaseManager, SyncManager
# contrib
import ujson as json
import numpy as np

# module tasktools
from tasktools.taskloop import coromask, renew, simple_fargs
from tasktools.assignator import TaskAssignator


# module networktools
from networktools.colorprint import gprint, bprint, rprint
from networktools.environment import get_env_variable
from networktools.ports import used_ports, get_port
from networktools.ssh import bridge, kill, clean_port
from networktools.library import pattern_value, \
    fill_pattern, context_split, \
    gns_loads, gns_dumps
from networktools.environment import get_env_variable

# module GNCSocket
# Implemente socket with this class better
# from gnsocket.server import GNSocketServer
from gnsocket.socket_server import GNCSocketServer
from gnsocket.socket_client import GNCSocketClient

# call settings variables
from settings import GUI_STATIONS_BY_WORKER, GUI_WORKERS, GROUP, st_dict
from settings import host, gnsocket_port, user, nproc
from settings import (RDB_SOURCE_HOST, RDB_ENU_HOST)
from settings import (RDB_SOURCE_PORT, RDB_ENU_PORT)
from settings import (COLLECTOR_SOCKET_IP, COLLECTOR_SOCKET_PORT)
from settings import (DATAWORK_SOCKET_IP, DATAWORK_SOCKET_PORT)

# call scheduler engine datawork
from datawork import DragonGather
from datawork import DataManager


host_server = socket.gethostbyname(socket.gethostname())

AEventLoop = type(asyncio.get_event_loop())

parser = argparse.ArgumentParser(description="Obtener parámetros de operación")

parser.add_argument('--group',
                    help="Select the group of stations {0,1, 2, 3, ALL}",)
parser.add_argument('--workers',
                    type=int,
                    help="Select amount of workers to run with")
parser.add_argument('--eXw',
                    type=int,
                    help="Amount of stations by workers, to get data and process")

if __name__ == "__main__":
    args = parser.parse_args()
    workers = nproc
    if args.workers:
        workers = args.workers
    est_by_proc = GUI_STATIONS_BY_WORKER
    group = GROUP
    if args.group:
        if args.group.isdigit():
            group = int(args.group)
            if group in st_dict.keys():
                group = st_dict[group]
        else:
            if args.group in st_dict.keys():
                group = st_dict[args.group]
    if args.eXw:
        est_by_proc = args.eXw

    dt_status = "GROUP"
    dt_group = group

    # create a bridge to collector system, to connect GNCSocket
    dw_server_address = (DATAWORK_SOCKET_IP, DATAWORK_SOCKET_PORT)
    clean_port(DATAWORK_SOCKET_PORT)
    tsleep = 1
    # GNCSocket port
    port = gnsocket_port
    up = used_ports()
    [local_port, up] = get_port(up)
    collector_address = (COLLECTOR_SOCKET_IP, COLLECTOR_SOCKET_PORT)
    dt_status = "ALL"
    uin_p = 4
    nproc = nproc
    loop = asyncio.get_event_loop()
    with concurrent.futures.ProcessPoolExecutor(workers) as executor:
        manager = Manager()
        sta_assigned = manager.dict()
        ipt = manager.list()
        ico = manager.list()
        # SOCKET SERVER network->terminal
        queue_n2t = manager.Queue()
        queue_t2n = manager.Queue()
        # SOCKET CLIETN network->terminal
        queue_client_n2t = manager.Queue()
        queue_client_t2n = manager.Queue()
        queue2client = manager.Queue()
        #
        queue_list = [queue_n2t, queue_t2n]
        # terminal->network
        process_queue = manager.Queue()
        stations_queue = manager.Queue()
        dbus_geojson_queue = manager.Queue()
        tasks_queue = manager.Queue()
        ans_queue = manager.Queue()
        dbus_queue = manager.Queue()
        proc_tasks = manager.dict()
        assigned_tasks = manager.dict()
        stations = manager.dict()
        position = manager.dict()
        db_data = manager.dict()
        sta_init = manager.dict()
        gui_group = manager.dict()

        kwargs = dict()
        status_tasks = manager.dict()

        # list of process identificators
        proc_tasks = manager.dict()
        signals = manager.dict()

        # create gui queue
        gui_queue = manager.Queue()
        gui_set = manager.list()

        # connect as client, available a server
        server_socket = GNCSocketServer(
            queue_n2t, queue_t2n, address=dw_server_address)
        print("Cliente ===================")
        print(collector_address)
        client_socket = GNCSocketClient(
            queue_client_n2t, queue_client_t2n, address=collector_address)
        enqueued = manager.list()

        kwargs.update({
            'enqueued': enqueued,
            'ipt': ipt,
            'ico': ico,
            'queue_client_n2t': queue_client_n2t,
            'queue_client_t2n': queue_client_t2n,
            'queue2client': queue2client,
            'stations': stations,
            'position': position,
            'db_data': db_data,
            'proc_tasks': proc_tasks,
            'nproc': nproc,
            'sta_init': sta_init,
            'bridge': bridge,
            'status_tasks': status_tasks,
            'assigned_tasks': assigned_tasks,
            'rdb_address': (RDB_SOURCE_HOST, RDB_SOURCE_PORT),
            'rdb_enu_address': (RDB_ENU_HOST, RDB_SOURCE_PORT),
            'rdb_dbname': 'collector',
            'chart': 'gnss_graph',
            'gui_set': gui_set,
            'gui_queue': gui_queue,
            'process_queue': process_queue,
            'signals': signals,
            'isg': manager.list(),
            'sigid': manager.list(),
            'group': dt_group,  # Specifi group
            'gui_group': gui_group,  # gui_group
            'dbus_queue': dbus_queue,
            'dbus_geojson_queue': dbus_geojson_queue,
            # if start collect data from rethindb is controled by gui
            'send_control': manager.Value(c_bool, True),
            'type_bus': 'session', })

        locker = manager.Lock()
        queue_ans_process = manager.Queue()
        # create new scheduler
        gather = DragonGather(queue_list, **kwargs)

        # assign scheduler yo assignator
        assignator = TaskAssignator(
            gather,
            process_queue,
            queue_ans_process,
            sta_assigned,
            dt_status,
            dt_group,
            enqueued,
            locker)

        mq_host = get_env_variable('MQ_HOST')
        mq_name = get_env_variable('MQ_NAME')
        mq_pass = get_env_variable('MQ_PASS')
        creds = (mq_name, mq_pass)
        mq_code = get_env_variable('MQ_CODE')
        mq_vhost = get_env_variable('MQ_VHOST')
        mq_queue_name = get_env_variable('MQ_QUEUE_NAME')

        """
        'code': 'gfast',
            'vhost': 'gfast',
            'host': mq_host,
            'credentials': creds,
            'qname': 'datawork2gfast',
        """
        amqp_data = {
            'amqp': True,
            'code': "enu_data",
            'vhost': mq_vhost,
            'host': mq_host,
            'credentials': creds,
            'queue_name': "geo_enu_data",
            'routing_key': 'data',
            'durable': False,
            'consumer_tag': 'enu_data'
        }
        amqp_data_eb = {
            'amqp': True,
            'code': "earlybird",
            'vhost': mq_vhost,
            'host': mq_host,
            'credentials': creds,
            'queue_name': "earlybird",
            'routing_key': 'earlybird',
            'exchange': 'earlybird',
            'durable': False,
            'consumer_tag': 'geojson'
        }
        amqp_status = {
            'amqp': True,
            'code': 'status',
            'vhost': mq_vhost,
            'host': mq_host,
            'credentials': creds,
            'queue_name': 'status_gnss',
            'routing_key': 'status',
            'exchange': 'status_exchange',
            'durable': True,
            'consumer_tag': 'status_data'
        }

        amqp_options = manager.dict(
            data=amqp_data,
            status=amqp_status,
            earlybird=amqp_data_eb
        )

        signal_geojson = DataManager(dbus_geojson_queue, amqp_options)

        tasks_list = []

        # signal task

        # second process -> tasks asignment

        signal_geojsontask = loop.run_in_executor(
            executor,
            signal_geojson.task_cycle)

        tasks_list.append(signal_geojsontask)

        # second process -> tasks asignment

        # Task on gather to process messages RPC
        # Active client SOCKET connected with Collector

        networktask = loop.run_in_executor(
            executor,
            gather.msg_network_task
        )
        tasks_list.append(networktask)

        # Task to receive on socket messages from outside
        try:
            socket_client_task = loop.run_in_executor(
                executor,
                client_socket.socket_task
            )
            pass
        except Exception as e:
            print("Falla en cargar socket task CLIENT")
            raise e

        tasks_list.append(socket_client_task)

        try:
            socket_task = loop.run_in_executor(
                executor,
                server_socket.socket_task
            )
            pass
        except Exception as e:
            print("Falla en cargar socket task SERVER")
            raise e

        tasks_list.append(socket_task)

        """

        socket_task ---> networktask
        A ---> B
        A sends to B using queues a message

        """
        NIPT = workers-5  # Workers avalaible

        bprint("###"*20)
        bprint("#=#"*20)
        print("TOTAL NIPT %s / workers %s" % (NIPT, workers))
        bprint("#=#"*20)
        bprint("###"*20)

        for i in range(NIPT):
            # time.sleep(1)
            ipt = gather.set_ipt(4)
            gather.proc_tasks[ipt] = []
            # problem here for python v3.6>
            task_process = loop.run_in_executor(
                executor,
                functools.partial(
                    gather.manage_tasks,
                    ipt))
            tasks_list.append(task_process)

        tasks_list.append(assignator.new_process_task())
        bprint("Iniciando tareas")
        loop.run_until_complete(asyncio.gather(*tasks_list))

    def manage_tasks(self, ipt):
        bprint(f"Init task process IPT {ipt}")
        return super().manage_task(ipt)



from station_codes import stations
from networktools.environment import get_env_variable
import multiprocessing

GUI_WORKERS = int(get_env_variable('GUI_WORKERS'))
GUI_STATIONS_BY_WORKER = int(get_env_variable('GUI_STATIONS_BY_WORKER'))

user = 'geodesia'
host = 'localhost'
#host = 'localhost'
gnsocket_port = 6677

nproc = multiprocessing.cpu_count()

SELECT_GROUP = int(get_env_variable('SELECT_GROUP'))

st_dict = dict([
    (0, ['TRPD',"CONZ"]),
    (1, ['AEDA', 'CHYT', 'CTPC', 'FMCO', 'JRGN', 'LSCH']),
    (2, ['MCLA', 'PATH', 'PAZU', 'PCCL', 'PFRJ', 'PVCA']),
    (3, ['QTAY', 'UDAT', 'UTAR', 'UAPE', 'QSCO', 'CGTC']),
    (4, ['QTAY', 'UDAT', 'UTAR', 'UAPE', 'QSCO', 'CGTC'])])

all_codes = stations

"""[
    'VALN',
    'HSCO',
    'QSCO',
    'FMCO',
    'PAZU',
    'JRGN',
    'PVCA',
    'CHYT',
    'UDAT',
    'NAVI',
    'PATH',
    'ATJN',
    'AEDA',
    'LSCH',
    'CCSN',
    'LLCH',
    'PCMU',
    'CRSC',
    'UAPE',
    'VLZL',
    'ZAPA',
    'QTAY',
    'PCCL',
    'PFRJ',
    'MCLA',
    'UTAR',
    'RCSD',
    'TRPD',
    'CGTC']
"""

#[all.extend(group) for group in st_dict.values()]

st_dict.update({"ALL": all_codes})

GROUP = st_dict.get(SELECT_GROUP, 0)
COLLECTOR_SOCKET_IP = get_env_variable('COLLECTOR_SOCKET_IP')
COLLECTOR_SOCKET_PORT = int(get_env_variable('COLLECTOR_SOCKET_PORT'))
print("Socket Port to ATLAS", COLLECTOR_SOCKET_IP, ":",COLLECTOR_SOCKET_PORT)
RDB_SOURCE_HOST = get_env_variable('RDB_SOURCE_HOST')
RDB_ENU_HOST = get_env_variable('RDB_ENU_HOST')
RDB_SOURCE = get_env_variable('RDB_SOURCE_PORT')
RDB_ENU = get_env_variable('RDB_ENU_PORT')
DATAWORK_SOCKET_IP = get_env_variable('DATAWORK_SOCKET_IP')
DATAWORK_SOCKET_PORT = int(get_env_variable('DATAWORK_SOCKET_PORT'))



import multiprocessing
import re
import json
from pathlib import Path
from networktools.environment import get_env_variable
from networktools.library import check_type
"""
Variables de ambiente para datawork
"""


class EnvData:
    def __init__(self, *args, **kwargs):
        self.LOG_PATH = get_env_variable("LOG_PATH")
        self.NPROC = multiprocessing.cpu_count()
        self.DW_DIR = Path(__file__).parent.parent.parent.resolve()
        self.DATAWORK_EST_X_PROC = get_env_variable("DATAWORK_EST_X_PROC")
        self.DATAWORK_TSLEEP = get_env_variable("DATAWORK_TSLEEP")
        self.DATAWORK_WORKERS = int(get_env_variable("DATAWORK_WORKERS"))
        self.DW_STATUS = get_env_variable("DW_STATUS")
        self.DW_GROUP = get_env_variable("DW_GROUP")
        self.SELECT_GROUP = get_env_variable("SELECT_GROUP")
        self.COLLECTOR_SOCKET_IP = get_env_variable("COLLECTOR_SOCKET_IP")
        self.COLLECTOR_SOCKET_PORT = get_env_variable("COLLECTOR_SOCKET_PORT")
        self.DATAWORK_SOCKET_IP = get_env_variable("DATAWORK_SOCKET_IP")
        self.DATAWORK_SOCKET_PORT = get_env_variable("DATAWORK_SOCKET_PORT")

        self.RDB_SOURCE_HOST = get_env_variable("RDB_SOURCE_HOST")
        self.RDB_SOURCE_PORT = get_env_variable("RDB_SOURCE_PORT")
        self.RDB_ENU_HOST = get_env_variable("RDB_ENU_HOST")
        self.RDB_ENU_PORT = get_env_variable("RDB_ENU_PORT")

        self.MQ_HOST = get_env_variable("MQ_HOST")
        # Es la dirección ip on url del host dónde se ubica el servicio
        self.MQ_NAME = get_env_variable("MQ_NAME")
        # Es el nombre de usuario que accede a la cola
        self.MQ_PASS = get_env_variable("MQ_PASS")
        # Es el la password asociada
        self.MQ_VHOST = get_env_variable("MQ_VHOST")
        # Es el nombre del host virtual de la cola
        self.MQ_CODE = get_env_variable("MQ_CODE")
        # Es el nombre en código de la cola
        self.MQ_QUEUE_NAME = get_env_variable("MQ_QUEUE_NAME")
        self.RDB_DBNAME = get_env_variable("RDB_DBNAME")

    def show(self):
        [print("export %s=%s" % (k, v)) for k, v in vars(self).items()]

    @property
    def json(self):
        return {k: str(v) for k, v in vars(self).items()}

    def save_json(self, name):
        json_file = re.compile("\.json$")
        if json_file.search(name):
            with open(name, "w") as f:
                json.dump(self.json, f, indent=2)
                f.write("\n")
        else:
            print("El nombre de archivo no es json, no se genera")


if __name__ == "__main__":
    envdata = EnvData()
    envdata.save_json("ejemplo_datawork.json")



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).resolve().parent

with open(this_directory/'README.md', encoding='utf-8') as f:
    readme = f.read()

with open(this_directory/'VERSION') as version_file:
    version = version_file.read().strip()

setup(name='datawork',
      version=version,
      description='Data Work system gather all data from a database to make analysis and charts',
      url='http://gitlab.csn.uchile.cl/dpineda/datawork',
      author='David Pineda Osorio',
      author_email='dpineda@csn.uchile.cl',
      license='GPLv3',
      packages=['datawork'],
      keywords=["collector", "gnss", "scheduler", "async", "multiprocess"],
      install_requires=["networktools",
                        "tasktools",
                        "basic_logtools",
                        "basic_queuetools",
                        "data_rdb",
                        "data_geo",
                        "data_amqp",
                        "gnsocket",
                        "uvloop",
                        "click",
                        "numpy",
                        "ujson"],
      entry_points={
          'console_scripts': ["datawork = datawork.scripts.run_datawork:run_datawork", ]
      },
      include_package_data=True,
      long_description=readme,
      long_description_content_type='text/markdown',
      zip_safe=False)
